import { Mapper } from './Mapper';

class Altitude {
  private migrations: Job[];
  private ready: Promise<any>;
  constructor(private mapper: Mapper) {
    let self = this;
    this.migrations = [];
    this.ready = new Promise((resolve, reject) => {
      this.shouldInitDB().then((shouldInit: boolean) => {
        if (shouldInit) {
          console.log(`Initializing ${this.mapper.database.connectionOptions.database}`);
          this.mapper.runQuery(`CREATE TABLE CHANGE_LOG (
            id int NOT NULL AUTO_INCREMENT,
            name VARCHAR(255),
            creator VARCHAR(255),
            UpdatedTime TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
            PRIMARY KEY (id)
          );`, []).then(() => {
            console.log(`Created CHANGE_LOG table.`);
            resolve(true);
          });
        } else {
          resolve(true);
        }
      }).catch((err) => {
        console.log(err)
      });
    });

  }

  public addMigration(newMigration: Job) {
    if (this.validateJob(newMigration)) {
      this.migrations.push(newMigration);
    } else {
      throw `The provided migration job (${newMigration.name}) is not valid.`;
    }
  }

  public removeMigration(migrationName: string) {
    this.migrations = this.migrations.filter((migration: Job) => {
      return migration.name !== migrationName;
    });
  }

  public runMigrationByName(jobName: string) {
    let migrationsToRun: Job[] = this.migrations.filter((job: Job) => {
      return job.name === jobName;
    });

    return this.runMigrations(migrationsToRun);
  }

  public runMigrations(migrationsToRun: Job[] = this.migrations) {
    let self = this;
    let defer = this.ready;
    migrationsToRun.forEach((job: Job) => {
      defer = defer.then((): Promise<any> => {
        return self.runMigration(job);
      });
    });
    return defer;
  }

  private validateJob(job: Job) {
    let valid = true;
    if (job.preCondition) {
      valid = valid && typeof job.preCondition === 'string';
    }
    if (job.preConditionFieldName) {
      valid = valid && typeof job.preConditionFieldName === 'string';
    }
    if (job.preConditionValidation) {
      valid = valid && typeof job.preConditionValidation === 'function';
    }
    if (job.postCondition) {
      valid = valid && typeof job.postCondition === 'string';
    }
    if (job.postConditionFieldName) {
      valid = valid && typeof job.postConditionFieldName === 'string';
    }
    if (job.postConditionValidation) {
      valid = valid && typeof job.postConditionValidation === 'function';
    }
    if (job.rollback) {
      valid = valid && typeof job.rollback === 'string';
    }
    valid = valid && typeof job.name === 'string';
    valid = valid && typeof job.query === 'string';
    valid = valid && typeof job.creator === 'string';

    return valid;
  }

  private shouldInitDB() {
    return this.mapper.runQuery(`SELECT table_name FROM information_schema.tables WHERE table_schema = '${this.mapper.database.connectionOptions.database}' AND table_name = 'CHANGE_LOG';`, [])
      .then((results: any[]) => {
        return results.length === 0;
      });
  }

  private alreadyRun(migration: Job) {
    return this.mapper.runQuery(`SELECT name FROM CHANGE_LOG WHERE name = '${migration.name}';`, [])
      .then((results: any[]) => {
        return results.length !== 0;
      });
  }

  private runMigration(migration: Job) {
    let self = this;
    return this.alreadyRun(migration)
      .then((alreadyRun: boolean) => {
        if (alreadyRun) {
          return `Skipping migration ${migration.name} because it has already been run.`;
        } else {
          return doMigration();
        }
      });

    function doMigration() {
      return self.runPreCondition(migration)
      .then((preConditionResult) => {
        if (preConditionResult) {
          let defer: Promise<any> = Promise.resolve();
          let queries: string[] = migration.query.split(';');
          queries = queries.filter((query) => query.length);
          queries.forEach((query: string) => {
            defer = defer.then((result) => {
              if (result) console.log(result);
              return self.mapper.runQuery(query, []);
            });
          });
          return defer;
        } else {
          throw 'PreCondition was not met.';
        }
      })
      .then((mainQueryResult) => {
        console.log(mainQueryResult);
        return self.runPostCondition(migration);
      })
      .then((postConditionResult) => {
        if (!postConditionResult) {
          return self.runRollback(migration);
        } else {
          return self.recordRun(migration)
          .then(() => {
            return 'Success';
          });
        }
      })
      .catch((error) => {
        console.log(error);
      });
    }
  }

  private recordRun(migration: Job) {
    return this.mapper.runQuery(`INSERT INTO CHANGE_LOG SET ?;`, [new ChangeLog(migration)]);
  }

  private runPreCondition(migration: Job) {
    let defer = new Promise((resolve, reject) => {
      if (migration.preCondition) {
        this.mapper.runQuery(migration.preCondition, []).then((result) => {
          console.log(result);
          let isValid = typeof migration.preConditionValidation === 'function' ? migration.preConditionValidation(result) : result.length > 0 && result[0][migration.preConditionFieldName || 'precondition'] !== 0;
          console.log(isValid);
          resolve(isValid);
        });
      } else {
        resolve(true);
      }
    });

    return defer;
  }

  private runPostCondition(migration: Job) {
    let defer = new Promise((resolve, reject) => {
      if (migration.postCondition) {
        this.mapper.runQuery(migration.postCondition, []).then((result) => {
          console.log(result);
          console.log(result[0][migration.postConditionFieldName || 'postcondition'] !== 0);
          resolve(result[0][migration.postConditionFieldName || 'postcondition'] !== 0);
        });
      } else {
        resolve(true);
      }
    });

    return defer;
  }

  private runRollback(migration: Job) {
    let defer = new Promise((resolve, reject) => {
      if (migration.rollback) {
        this.mapper.runQuery(migration.rollback, []).then((result) => {
          resolve(result);
        }).catch((error) => {
          reject(error);
        });
      } else {
        resolve(true);
      }
    });

    return defer;
  }
}

interface ValidationFunction {
  (results: any[]): boolean;
}

interface Job {
  preCondition?: string;
  preConditionFieldName?: string;
  preConditionValidation?: ValidationFunction;
  postCondition?: string;
  postConditionFieldName?: string;
  postConditionValidation?: ValidationFunction;
  rollback?: string;
  name: string;
  query: string;
  creator: string;
}

class ChangeLog {
  public name: string;
  public creator: string;

  constructor(migration: Job) {
    this.name = migration.name;
    this.creator = migration.creator;
  }
}

export {
  Altitude,
  Job
}
