import {ColumnDef, ColumnRef, ComponentDef, EndpointDef, Filter, SystemDef, TableDef} from "../systemGenService";
import path from "path";
import fs from "fs";
import {createDetailsInitValues, findColumn, makeSet, uppercaseFirstLetter} from "../helpers";
import pluralize from "pluralize";
import {getFuncParams} from "../views/service-creator";

const INIT_VALUE = {
    "blob": "''",
    "boolean": "false",
    "date": "''",
    "dateTIME": "''",
    "number": "0",
    "string": "''"
};

export function buildListView(component: ComponentDef, outDir: string, systemDef: SystemDef): Promise<void> {
    let hasDeleteEndpoint = false;
    let hasItemEndpoint = false;
    let hasUpdateEndpoint = false;
    let hasCreateEndpoint = false;
    let listEndpoints: EndpointDef[] = component.endpoints.filter((epd: EndpointDef) => {
        if (epd.type === 'delete') {
            hasDeleteEndpoint = true;
        }
        if (epd.type === 'create') {
            hasCreateEndpoint = true;
        }
        if (epd.type === 'update') {
            hasUpdateEndpoint = true;
        }
        if (epd.type === 'item') {
            hasItemEndpoint = true;
        }
        return epd.type === 'list' || epd.type === 'search';
    });
    let filterParams: Filter[] = [];
    let endpointPromise = Promise.resolve();
    for (let i in listEndpoints) {
        if (i === '0') {
            insertFEPropsCode(listEndpoints[i], outDir, systemDef);
        }
        if (listEndpoints[i].filters) {
            filterParams = [...filterParams, ...listEndpoints[i].filters as Filter[]];
        }
        endpointPromise = endpointPromise.then(() => {
            return insertFEListCode(listEndpoints[i], outDir, systemDef, hasDeleteEndpoint);
        });
    }
    if (listEndpoints.length) {
        endpointPromise = endpointPromise.then(() => {
            return insertFEListMarkup(listEndpoints[0], outDir, hasDeleteEndpoint)
        });
    }
    if (hasDeleteEndpoint && listEndpoints.length) {
        endpointPromise = endpointPromise.then(() => {
           return insertFEDeleteCode(listEndpoints[0], outDir);
        });
    }
    // combine all duplicate filterParams
    // @ts-ignore
    let filterSet: Filter[] = makeSet<Filter>(filterParams, 'param');
    // add filterParams to the markup as filter options
    if (listEndpoints.length) {
        endpointPromise = endpointPromise.then(() => {
            return insertFEFiltersMarkup(listEndpoints[0], filterSet, outDir);
        });
    }

    // TODO: add create component to sidebar
    // TODO: add details component to sidebar
    // TODO: if delete endpoint exists add delete button next to each row

    return endpointPromise;
}

function insertFEListCode(view: EndpointDef, outDir: string, systemDef: SystemDef, hasDeleteEndpoint: boolean) {
    return new Promise<void>((resolve) => {
        const separator: string = `// SYSTEM-BUILDER-component-functions`;
        let fileLocation = path.join(outDir, `${view.component}List.vue`)
        let initServiceFile: string = fs.readFileSync(fileLocation, 'utf8');
        let parts = initServiceFile.split(separator);
        parts[0] = parts[0] + createListRequestFunc(view, systemDef, hasDeleteEndpoint);
        let newServiceFile = parts.join(separator);
        fs.writeFileSync(fileLocation, newServiceFile, 'utf8');
        resolve();
    });
}

function insertFEDeleteCode(view: EndpointDef, outDir: string) {
    return new Promise<void>((resolve) => {
        const separator: string = `// SYSTEM-BUILDER-component-functions`;
        let fileLocation = path.join(outDir, `${view.component}List.vue`)
        let initServiceFile: string = fs.readFileSync(fileLocation, 'utf8');
        let parts = initServiceFile.split(separator);
        parts[0] = parts[0] + createFEDeleteCode(view);
        let newServiceFile = parts.join(separator);
        fs.writeFileSync(fileLocation, newServiceFile, 'utf8');
        resolve();
    });
}

function insertFEListMarkup(view: EndpointDef, outDir: string, hasDeleteEndpoint: boolean) {
    return new Promise<void>((resolve) => {
        const separator: string = `<!-- SYSTEM-BUILDER-item-props -->`;
        let fileLocation = path.join(outDir, `${view.component}List.vue`)
        let initServiceFile: string = fs.readFileSync(fileLocation, 'utf8');
        let parts = initServiceFile.split(separator);
        parts[0] = parts[0] + createListMarkup(view, hasDeleteEndpoint);
        let newServiceFile = parts.join(separator);
        fs.writeFileSync(fileLocation, newServiceFile, 'utf8');
        resolve();
    });
}

function insertFEFiltersMarkup(view: EndpointDef, filters: Filter[], outDir: string) {
    return new Promise<void>((resolve) => {
        const separator: string = `<!-- SYSTEM-BUILDER-filter-options -->`;
        let fileLocation = path.join(outDir, `${view.component}List.vue`)
        let initServiceFile: string = fs.readFileSync(fileLocation, 'utf8');
        let parts = initServiceFile.split(separator);
        parts[0] = parts[0] + createFilterMarkup(view, filters);
        let newServiceFile = parts.join(separator);
        fs.writeFileSync(fileLocation, newServiceFile, 'utf8');
        resolve();
    });
}

function insertFEPropsCode(view: EndpointDef, outDir: string, systemDef: SystemDef) {
    return new Promise<void>((resolve) => {
        const separator: string = `// SYSTEM-BUILDER-init-props`;
        let fileLocation = path.join(outDir, `${view.component}List.vue`)
        let initServiceFile: string = fs.readFileSync(fileLocation, 'utf8');
        let parts = initServiceFile.split(separator);
        parts[0] = parts[0] + createDetailsInitValues(view, systemDef);
        let newServiceFile = parts.join(separator);
        fs.writeFileSync(fileLocation, newServiceFile, 'utf8');
        resolve();
    });
}

function createFilterMarkup(view: EndpointDef, filters: Filter[]): string {
    let out = ``;
    let inputType: {[key: string]: string} = {
        "blob": "text",
        "boolean": "checkbox",
        "date": "date",
        "dateTIME": "datetime-local",
        "number": "number",
        "string": "text",
    };
    filters.forEach((f: Filter) => {
        if (f.type) {
            // base the filter markup on type
            out = out + `<div>
            <label>${f.param}: </label>
            <input type="${inputType[f.type] || 'text'}" placeholder="${f.param}" v-model="filterItems.${f.param}"/>
        </div>
        `;
        } else {
            // default to a string filter
            out = out + `<div>
            <input type="text" placeholder="${f.param}" v-model="filterItems.${f.param}"/>
        </div>
    `;
        }
    });
    out = out + `<div>
            <button @click="search${uppercaseFirstLetter(pluralize(view.component))}()">Search</button>
        </div>
    `;
    return out;
}

function createListMarkup(view: EndpointDef, hasDeleteEndpoint: boolean): string {
    let out = ``;

    view.columns.forEach((column: ColumnRef) => {
       out = out + `<div class="">{{ item.${column.name} }}</div>
      `;
    });
    if (hasDeleteEndpoint) {
        out = out + `<div class="red-text" @click="deleteItem(item.${view.component}_id)">delete</div>
      `;
    }

    return out;
}

function createListRequestFunc(view: EndpointDef, systemDef: SystemDef, hasDeleteEndpoint: boolean) {
    let isSearch: boolean = view.type === 'search';
    let params: string[] = getFuncParams(view);
    params = params.map((param: string) => {
        return 'this.filterItems.' + param;
    });
    params.push('0'); // offset
    params.push('1000'); // limit

    let out = `private ${isSearch ? 'search' : 'get'}${uppercaseFirstLetter(pluralize(view.component))}() {
    this.loading = true;
    this.errorMessage = '';
    ${view.component}Service.${isSearch ? 'search' : 'get'}${uppercaseFirstLetter(pluralize(view.component))}(${params.join(', ')}).then((listItems: ${uppercaseFirstLetter(view.component)}Config[]) => {
        // success
        this.loading = false;
        this.errorMessage = '';
        this.items = listItems;
    }).catch(this.handleError.bind(this));
  }

  `;
    return out;
}

function createFEDeleteCode(view: EndpointDef) {
    let out = `private deleteItem(itemKey: string) {
    this.loading = true;
    this.errorMessage = '';
    ${view.component}Service.delete${uppercaseFirstLetter(view.component)}(itemKey).then(() => {
        this.loading = false;
        this.errorMessage = '';
        return this.get${uppercaseFirstLetter(pluralize(view.component))}();
    }).catch(this.handleError.bind(this));  
  }
        
  `;
    return out;
}

