import {ColumnRef, EndpointDef, Filter, JoinDef, Order} from "../systemGenService";
import pluralize from "pluralize";
import {uppercaseFirstLetter, lowercaseFirstLetter} from "../helpers";

function createMapperFunc(view: EndpointDef): string {
    let func: string = '';
    if (view.type === 'list' || view.type === 'count') {
        func = buildGetList(view);
    } else if (view.type === 'search') {
        func = buildGetList(view, true);
    } else if (view.type === 'item') {
        func = buildGetItem(view);
    } else if (view.type === 'update') {
        func = buildUpdateFunc(view);
    } else if (view.type === 'create') {
        func = buildCreateFunc(view);
    } else if (view.type === 'delete') {
        func = buildDeleteItem(view);
    }
    return func;
}

function buildDeleteItem(view: EndpointDef): string {
    let func: string = '';
    let funcName = buildMapperFunctionName(view);
    let tables: string = `${view.table}`;
    let {filters, requiredQueryValues, optionalFilterParts} = constructFilters(view);
    let query: string = `DELETE FROM ${tables}${filters}`;
    let funcParams = 'params: {[key: string]: string}';

    func = `
  public ${funcName}(${funcParams}) {
    
    let query: string = '${query}';
    let queryValues: (string | number)[] = [${requiredQueryValues}];
    
    // optional params?
    ${optionalFilterParts}
    return super.runQuery(query, [...queryValues]);
  }
  `;

    return func;
}

function buildGetItem(view: EndpointDef): string {
    let func: string = '';
    let funcName = buildMapperFunctionName(view);
    let columns: string = `${view.component}.${view.component}_id, ${view.columns.map(c => `${c.table}.${c.name}`).join(', ')}`;
    let tables: string = view.join?.length ? buildJoin(view) : `${view.table}`;
    let {filters, requiredQueryValues, optionalFilterParts} = constructFilters(view);
    let query: string = `SELECT ${columns} FROM ${tables}${filters} LIMIT 1`;
    let funcParams = 'params: {[key: string]: string}';

    func = `
  public ${funcName}(${funcParams}) {
    
    let query: string = '${query}';
    let queryValues: (string | number)[] = [${requiredQueryValues}];
    
    // optional params?
    ${optionalFilterParts}
    return super.runQuery(query, [...queryValues]);
  }
`;

    return func;
}

function buildCreateFunc(view: EndpointDef): string {
    let func: string = '';
    let inQuery: string = `INSERT INTO ${view.table} SET `;
    let {funcName, query, requiredQueryValues, optionalValuesCode} = getCreateOrUpdateFunc(inQuery, view);

    func = `
  public ${funcName}(params: {[key: string]: string}) {
    
    let query: string = '${query}';
    let queryValues: (string | number)[] = [${requiredQueryValues}];
    
    // optional params?
    ${optionalValuesCode}
    return super.runQuery(query, [...queryValues]);
  }
`;
    return func;
}

function buildUpdateFunc(view: EndpointDef): string {
    let func: string = '';
    let inQuery: string = `UPDATE ${view.table} SET `;
    let {funcName, query, requiredQueryValues, optionalValuesCode} = getCreateOrUpdateFunc(inQuery, view);
    let postQuery = ` WHERE `;
    for (let i in view.filters) {
        // @ts-ignore
        let filter: Filter = view.filters[i];
        if (!postQuery.endsWith('WHERE ')) {
            postQuery = postQuery + ` AND `;
        }
        postQuery = postQuery + `${filter.column.table}.${filter.column.name} = ?`;
        if (requiredQueryValues.length) {
            requiredQueryValues = requiredQueryValues + `, `;
        }
        requiredQueryValues = requiredQueryValues + `params.${filter.param}`;
    }

    func = `
  public ${funcName}(params: {[key: string]: string}) {
    
    let query: string = '${query}';
    let queryValues: (string | number)[] = [${requiredQueryValues}];
    
    // optional params?
    ${optionalValuesCode}
    
    query = query + '${postQuery}';
    return super.runQuery(query, [...queryValues]);
  }
`;

    return func;
}

function getCreateOrUpdateFunc(query: string, view: EndpointDef) {
    let funcName = buildMapperFunctionName(view);
    let requiredQueryValues: string = '';
    let optionalValuesCode: string = '';
    for (let i in view.columns) {
        let c: ColumnRef = view.columns[i];
        if (c.required) {
            if (!query.endsWith('SET ')) {
                query = query + `, `;
            }
            query = query + `${c.table}.${c.name} = ?`;
            if (requiredQueryValues.length) {
                requiredQueryValues = requiredQueryValues + `, `;
            }
            requiredQueryValues = requiredQueryValues + `params.${c.param}`;
        } else {
            optionalValuesCode = optionalValuesCode + `
    if (params.${c.param}) {
      query = query + ', ${c.table}.${c.name} = ?';
      queryValues.push(params.${c.param});
    }
`;
        }
    }
    return {funcName, query, requiredQueryValues, optionalValuesCode};
}

function constructFilters(view: EndpointDef, isSearch=false) {
    let filters: string = '';
    let requiredFilterStrings = [];
    let requiredQueryValues = '';
    let optionalFilterParts = '';
    if (view.filters) {
        for (let i in view.filters) {
            let f: Filter = view.filters[i];
            if (f.required) {
                requiredFilterStrings.push(`${f.column.table}.${f.column.name} ${f.comparison} ?`); // e.g. 'table.id = ?'
                if (requiredQueryValues.length) {
                    requiredQueryValues = requiredQueryValues + `, `;
                }
                requiredQueryValues = requiredQueryValues + `params.${f.param}`;
            } else {
                /* e.g.
                  if (eventLog.app) {
                    query = query + ', app = ?';
                    queryValues.push(eventLog.app);
                  }
                 */
                optionalFilterParts = optionalFilterParts + `
    if (params.${f.param}) {
      query = query + ' ${isSearch ? 'OR' : 'AND'} ${f.column.table}.${f.column.name} ${f.comparison} ?';
      queryValues.push(${isSearch ? '\'%\' + params.' + f.param + ' + \'%\'' : 'params.' + f.param});
    }`;
            }
        }
        filters = ` WHERE ${requiredFilterStrings.join(' AND ')}`;
    }
    return {filters, requiredQueryValues, optionalFilterParts};
}

function buildGetList(view: EndpointDef, isSearch=false): string {
    // TODO: optionally allow a raw query definition with predefined params
    // This would be so much simpler and easy if it were just the queries wanted
    let func: string = '';
    let funcName: string = '';
    let columns: string = view.type === 'count' ? `count(*) as Count` : `${view.component}.${view.component}_id, ${view.columns.map(c => `${c.table}.${c.name}`).join(', ')}`;
    let query: string = '';
    let tables: string = view.join?.length ? buildJoin(view) : `${view.table}`;
    funcName = buildMapperFunctionName(view);

    let {filters, requiredQueryValues, optionalFilterParts} = constructFilters(view, isSearch);

    query = `SELECT ${columns} FROM ${tables}${filters}`;
    let orderByCode = '';
    if (view.orderBy) {
        orderByCode = `
    query = query + '${buildOrderBy(view)}';
`;
    }
    let funcParams = 'params: {[key: string]: string}, offset: number, limit: number';
    let limitOrderCode = `${orderByCode}
    query = query + ' LIMIT ?, ?';
    queryValues.push(offset);
    queryValues.push(limit);`;
    if (view.type === 'count') {
        funcParams = 'params: {[key: string]: string}';
        limitOrderCode = ``;
    }

    func = `
  public ${funcName}(${funcParams}) {
    
    let query: string = '${query}';
    let queryValues: (string | number)[] = [${requiredQueryValues}];
    
    // optional params?
    ${optionalFilterParts}
    ${limitOrderCode}
    return super.runQuery(query, [...queryValues]);
  }
`;
    return func;
}

function buildJoin(view: EndpointDef): string {
    if (view.join?.length) {
        let join = `${view.table}`;
        for (let i in view.join) {
            let j: JoinDef = view.join[i];
            join = join + ` JOIN ${j.table} ON ${j.on.left.table}.${j.on.left.name} = ${j.on.right.table}.${j.on.right.name}`;
        }
        return join;
    }
    return '';
}

function buildOrderBy(view: EndpointDef): string {
    let orderBy: string = '';
    if (view.orderBy?.length) {
        orderBy = orderBy + ' ORDER BY ';
        for (let i in view.orderBy) {
            let order: Order = view.orderBy[i];
            if (parseInt(i) !== 0) {
                orderBy = orderBy + ', ';
            }
            orderBy = orderBy + `${order.column.table}.${order.column.name} ${order.direction}`;
        }
    }
    return orderBy;
}

function buildMapperFunctionName(view: EndpointDef): string {
    let selector: {[type: string]: (view: EndpointDef) => string} = {
        'list': (view: EndpointDef) => `get${pluralize(uppercaseFirstLetter(view.component))}` + (view.type === 'count' ? 'Count' : ''),
        'count': (view: EndpointDef) => `get${pluralize(uppercaseFirstLetter(view.component))}` + (view.type === 'count' ? 'Count' : ''),
        'item': (view: EndpointDef) => `get${uppercaseFirstLetter(view.component)}`,
        'update': (view: EndpointDef) => `update${uppercaseFirstLetter(view.component)}`,
        'create': (view: EndpointDef) => `create${uppercaseFirstLetter(view.component)}`,
        'delete': (view: EndpointDef) => `delete${uppercaseFirstLetter(view.component)}`,
        'search': (view: EndpointDef) => `search${pluralize(uppercaseFirstLetter(view.component))}`,
    };
    return selector[view.type](view);
}

export {
    createMapperFunc,
    buildMapperFunctionName
}